@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-#  Copyright 2019, 2020 Ramil Nugmanov <nougmanoff@protonmail.com>
+#  Copyright 2019-2021 Ramil Nugmanov <nougmanoff@protonmail.com>
 #  This file is part of CGRtools.
 #
 #  CGRtools is free software; you can redistribute it and/or modify
@@ -19,10 +19,15 @@
 from CachedMethods import cached_property
 from collections import defaultdict, deque
 from logging import info
-from typing import Dict, Optional, Set, Tuple, Union
+from typing import Dict, Optional, Set, Tuple, Union, TYPE_CHECKING
 from ..exceptions import AtomNotFound, IsChiral, NotChiral
 
 
+if TYPE_CHECKING:
+    from CGRtools import MoleculeContainer, QueryContainer
+    Container = Union[MoleculeContainer, QueryContainer]
+
+
 def _pyramid_sign(n, u, v, w):
     #
     #  |   n /
@@ -126,7 +131,7 @@ def _allene_sign(n, u, v, w):
 class Stereo:
     __slots__ = ()
 
-    def clean_stereo(self):
+    def clean_stereo(self: 'Container'):
         """
         Remove stereo data
         """
@@ -135,7 +140,7 @@ class Stereo:
         self._cis_trans_stereo.clear()
         self.flush_cache()
 
-    def get_mapping(self, other, **kwargs):
+    def get_mapping(self: 'Container', other: 'Container', **kwargs):
         atoms_stereo = self._atoms_stereo
         allenes_stereo = self._allenes_stereo
         cis_trans_stereo = self._cis_trans_stereo
@@ -188,63 +193,7 @@ class Stereo:
         else:
             yield from super().get_mapping(other, **kwargs)
 
-    @cached_property
-    def _wedge_map(self):
-        plane = self._plane
-        atoms_stereo = self._atoms_stereo
-        allenes_centers = self._stereo_allenes_centers
-        atoms = self._atoms
-        used = set()
-        wedge = []
-        for n, s in self._allenes_stereo.items():
-            env = self._stereo_allenes[n]
-            term = self._stereo_allenes_terminals[n]
-            order = [(*env[:2], *term), (*env[1::-1], *term[::-1])]
-            if env[2]:
-                order.append((env[2], env[1], *term))
-                order.append((env[1], env[2], *term[::-1]))
-            if env[3]:
-                order.append((env[3], env[0], *term[::-1]))
-                order.append((env[0], env[3], *term))
-            order = sorted(order, key=lambda x: (x[0] in atoms_stereo, x[0] in allenes_centers,
-                                                 -atoms[x[0]].atomic_number))
-            while (order[0][0], order[0][2]) in used:
-                order.append(order.pop(0))
-            order = order[0]
-            used.add((order[2], order[0]))
-            s = self._translate_allene_sign(n, *order[:2])
-            v = _allene_sign((*plane[order[0]], 1), plane[order[2]], plane[order[3]], (*plane[order[1]], 0))
-            if not v:
-                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
-            if s:
-                wedge.append((order[2], order[0], v))
-            else:
-                wedge.append((order[2], order[0], -v))
-
-        for n, s in atoms_stereo.items():
-            order = sorted(self._stereo_tetrahedrons[n], key=lambda x: (x in atoms_stereo, x in allenes_centers,
-                                                                        -atoms[x].atomic_number, atoms[x].in_ring))
-            while (order[0], n) in used:
-                order.append(order.pop(0))
-            used.add((n, order[0]))
-
-            s = self._translate_tetrahedron_sign(n, order)
-            # need recalculation if XY changed
-            if len(order) == 3:
-                v = _pyramid_sign((*plane[n], 0),
-                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
-            else:
-                v = _pyramid_sign((*plane[order[3]], 0),
-                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
-            if not v:
-                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
-            if s:
-                wedge.append((n, order[0], v))
-            else:
-                wedge.append((n, order[0], -v))
-        return tuple(wedge)
-
-    def _translate_tetrahedron_sign(self, n, env):
+    def _translate_tetrahedron_sign(self: 'Container', n, env):
         """
         Get sign of chiral tetrahedron atom for specified neighbors order
 
@@ -252,9 +201,9 @@ class Stereo:
         :param env: neighbors order
         """
         s = self._atoms_stereo[n]
-        return self._translate_tetrahedron_sign_reversed(n , env, s)
+        return self._translate_tetrahedron_sign_reversed(n, env, s)
 
-    def _translate_tetrahedron_sign_reversed(self, n, env, s):
+    def _translate_tetrahedron_sign_reversed(self: 'Container', n, env, s):
         order = self._stereo_tetrahedrons[n]
         if len(order) == 3:
             if len(env) == 4:  # hydrogen atom passed to env
@@ -274,7 +223,7 @@ class Stereo:
             return not s
         return s
 
-    def _translate_cis_trans_sign(self, n, m, nn, nm):
+    def _translate_cis_trans_sign(self: 'Container', n, m, nn, nm):
         """
         Get sign for specified opposite neighbors
 
@@ -291,7 +240,7 @@ class Stereo:
             nn, nm = nm, nn
         return self._translate_cis_trans_sign_reversed(n, m, nn, nm, s)
 
-    def _translate_cis_trans_sign_reversed(self, n, m, nn, nm, s):
+    def _translate_cis_trans_sign_reversed(self: 'Container', n, m, nn, nm, s):
         atoms = self._atoms
         n0, n1, n2, n3 = self._stereo_cis_trans[(n, m)]
         if nn == n0:  # same start
@@ -333,7 +282,7 @@ class Stereo:
             return not s
         return s
 
-    def _translate_allene_sign(self, c, nn, nm):
+    def _translate_allene_sign(self: 'Container', c, nn, nm):
         """
         get sign for specified opposite neighbors
 
@@ -344,7 +293,7 @@ class Stereo:
         s = self._allenes_stereo[c]
         return self._translate_allene_sign_reversed(c, nn, nm, s)
 
-    def _translate_allene_sign_reversed(self, c, nn, nm, s):
+    def _translate_allene_sign_reversed(self: 'Container', c, nn, nm, s):
         atoms = self._atoms
 
         n0, n1, n2, n3 = self._stereo_allenes[c]
@@ -388,7 +337,7 @@ class Stereo:
         return s
 
     @cached_property
-    def _stereo_cumulenes(self) -> Dict[Tuple[int, ...], Tuple[int, int, Optional[int], Optional[int]]]:
+    def _stereo_cumulenes(self: 'Container') -> Dict[Tuple[int, ...], Tuple[int, int, Optional[int], Optional[int]]]:
         """
         Cumulenes which contains at least one non-hydrogen neighbor on both ends
         """
@@ -401,9 +350,14 @@ class Stereo:
         atoms = self._atoms
         cumulenes = {}
         for path in self.cumulenes:
+            nf = bonds[path[0]]
+            nl = bonds[path[-1]]
             n1, m1 = path[1], path[-2]
-            nn = [x for x in bonds[path[0]] if x != n1 and atoms[x].atomic_number != 1]
-            mn = [x for x in bonds[path[-1]] if x != m1 and atoms[x].atomic_number != 1]
+            if any(b not in (1, 4) for m, b in nf.items() if m != n1) or \
+                    any(b not in (1, 4) for m, b in nl.items() if m != m1):
+                continue
+            nn = [x for x in nf if x != n1 and atoms[x].atomic_number != 1]
+            mn = [x for x in nl if x != m1 and atoms[x].atomic_number != 1]
             if nn and mn:
                 sn = nn[1] if len(nn) == 2 else None
                 sm = mn[1] if len(mn) == 2 else None
@@ -411,7 +365,7 @@ class Stereo:
         return cumulenes
 
     @cached_property
-    def _stereo_tetrahedrons(self) -> Dict[int, Union[Tuple[int, int, int], Tuple[int, int, int, int]]]:
+    def _stereo_tetrahedrons(self: 'Container') -> Dict[int, Union[Tuple[int, int, int], Tuple[int, int, int, int]]]:
         """
         Tetrahedrons which contains at least 3 non-hydrogen neighbors
         """
@@ -483,7 +437,7 @@ class Stereo:
 class MoleculeStereo(Stereo):
     __slots__ = ()
 
-    def add_wedge(self, n: int, m: int, mark: bool, *, clean_cache=True):
+    def add_wedge(self: 'MoleculeContainer', n: int, m: int, mark: bool, *, clean_cache=True):
         """
         Add stereo data by wedge notation of bonds. Use it for tetrahedrons of allenes.
 
@@ -547,7 +501,7 @@ class MoleculeStereo(Stereo):
                 # only tetrahedrons and allenes supported
                 raise NotChiral
 
-    def calculate_cis_trans_from_2d(self, *, clean_cache=True):
+    def calculate_cis_trans_from_2d(self: 'MoleculeContainer', *, clean_cache=True):
         """
         Calculate cis-trans stereo bonds from given 2d coordinates. Unusable for SMILES and INCHI.
         """
@@ -571,7 +525,7 @@ class MoleculeStereo(Stereo):
         if flag and clean_cache:
             self.flush_cache()
 
-    def add_atom_stereo(self, n: int, env: Tuple[int, ...], mark: bool, *, clean_cache=True):
+    def add_atom_stereo(self: 'MoleculeContainer', n: int, env: Tuple[int, ...], mark: bool, *, clean_cache=True):
         """
         Add stereo data for specified neighbors bypass. Use it for tetrahedrons of allenes.
 
@@ -599,7 +553,8 @@ class MoleculeStereo(Stereo):
         else:  # only tetrahedrons supported
             raise NotChiral
 
-    def add_cis_trans_stereo(self, n: int, m: int, n1: int, n2: int, mark: bool, *, clean_cache=True):
+    def add_cis_trans_stereo(self: 'MoleculeContainer', n: int, m: int, n1: int, n2: int, mark: bool, *,
+                             clean_cache=True):
         """
         Add stereo data to cis-trans double bonds (not allenes).
 
@@ -632,7 +587,7 @@ class MoleculeStereo(Stereo):
         else:
             raise NotChiral
 
-    def _fix_stereo(self):
+    def _fix_stereo(self: 'MoleculeContainer'):
         if self._atoms_stereo:  # filter tetrahedrons
             stereo_tetrahedrons = self._stereo_tetrahedrons
             atoms_stereo = {k: v for k, v in self._atoms_stereo.items() if k in stereo_tetrahedrons}
@@ -691,6 +646,62 @@ class MoleculeStereo(Stereo):
             # flush cache
             del self.__dict__['_MoleculeStereo__chiral_centers']
 
+    @cached_property
+    def _wedge_map(self: 'Container'):
+        plane = self._plane
+        atoms_stereo = self._atoms_stereo
+        allenes_centers = self._stereo_allenes_centers
+        atoms = self._atoms
+        used = set()
+        wedge = []
+        for n, s in self._allenes_stereo.items():
+            env = self._stereo_allenes[n]
+            term = self._stereo_allenes_terminals[n]
+            order = [(*env[:2], *term), (*env[1::-1], *term[::-1])]
+            if env[2]:
+                order.append((env[2], env[1], *term))
+                order.append((env[1], env[2], *term[::-1]))
+            if env[3]:
+                order.append((env[3], env[0], *term[::-1]))
+                order.append((env[0], env[3], *term))
+            order = sorted(order, key=lambda x: (x[0] in atoms_stereo, x[0] in allenes_centers,
+                                                 -atoms[x[0]].atomic_number))
+            while (order[0][0], order[0][2]) in used:
+                order.append(order.pop(0))
+            order = order[0]
+            used.add((order[2], order[0]))
+            s = self._translate_allene_sign(n, *order[:2])
+            v = _allene_sign((*plane[order[0]], 1), plane[order[2]], plane[order[3]], (*plane[order[1]], 0))
+            if not v:
+                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
+            if s:
+                wedge.append((order[2], order[0], v))
+            else:
+                wedge.append((order[2], order[0], -v))
+
+        for n, s in atoms_stereo.items():
+            order = sorted(self._stereo_tetrahedrons[n], key=lambda x: (x in atoms_stereo, x in allenes_centers,
+                                                                        -atoms[x].atomic_number, atoms[x].in_ring))
+            while (order[0], n) in used:
+                order.append(order.pop(0))
+            used.add((n, order[0]))
+
+            s = self._translate_tetrahedron_sign(n, order)
+            # need recalculation if XY changed
+            if len(order) == 3:
+                v = _pyramid_sign((*plane[n], 0),
+                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
+            else:
+                v = _pyramid_sign((*plane[order[3]], 0),
+                                  (*plane[order[0]], 1), (*plane[order[1]], 0), (*plane[order[2]], 0))
+            if not v:
+                info(f'need 2d clean. wedge stereo ambiguous for atom {{{n}}}')
+            if s:
+                wedge.append((n, order[0], v))
+            else:
+                wedge.append((n, order[0], -v))
+        return tuple(wedge)
+
     @property
     def _chiral_tetrahedrons(self) -> Set[int]:
         return self.__chiral_centers[0]
@@ -703,8 +714,14 @@ class MoleculeStereo(Stereo):
     def _chiral_allenes(self) -> Set[int]:
         return self.__chiral_centers[2]
 
+    @property
+    def _chiral_morgan(self: Union['MoleculeContainer', 'MoleculeStereo']) -> Dict[int, int]:
+        if self._atoms_stereo or self._allenes_stereo or self._cis_trans_stereo:
+            return self.__chiral_centers[3]
+        return self.atoms_order
+
     @cached_property
-    def _stereo_axises(self) -> Tuple[Tuple[Tuple[int, ...], ...], Tuple[Tuple[int, ...], ...]]:
+    def _stereo_axises(self: 'MoleculeContainer') -> Tuple[Tuple[Tuple[int, ...], ...], Tuple[Tuple[int, ...], ...]]:
         """
         Get all stereogenic axises in rings with attached cumulenes. Stereogenic axises has only stereogenic atoms.
         """
@@ -809,7 +826,7 @@ class MoleculeStereo(Stereo):
         return tuple(out), tuple(env)
 
     @cached_property
-    def __stereo_axises(self):
+    def __stereo_axises(self: 'MoleculeContainer'):
         bonds = self._bonds
         tetrahedrons = self._stereo_tetrahedrons
         cumulenes = self._stereo_cumulenes
@@ -842,7 +859,7 @@ class MoleculeStereo(Stereo):
         return axises
 
     @cached_property
-    def __chiral_centers(self):
+    def __chiral_centers(self: Union['MoleculeContainer', 'MoleculeStereo']):
         atoms_stereo = self._atoms_stereo
         cis_trans_stereo = self._cis_trans_stereo
         allenes_stereo = self._allenes_stereo
@@ -963,14 +980,7 @@ class MoleculeStereo(Stereo):
                 morgan = self._morgan({**morgan, **morgan_update})
                 morgan_update = {}
             else:
-                return chiral_t, {(n, m) for n, *_, m in chiral_c}, {path[len(path) // 2] for path in chiral_a}
-
-
-class QueryStereo(Stereo):  # todo: implement add_wedge
-    __slots__ = ()
-
-    def add_wedge(self, n: int, m: int, mark: bool, *, clean_cache=True):
-        raise NotImplementedError
+                return chiral_t, {(n, m) for n, *_, m in chiral_c}, {path[len(path) // 2] for path in chiral_a}, morgan
 
 
 # 1  2
@@ -997,4 +1007,4 @@ _alkene_translate = {(0, 1): False, (1, 0): False, (0, 3): True, (3, 0): True,
                      (2, 3): False, (3, 2): False, (2, 1): True, (1, 2): True}
 
 
-__all__ = ['MoleculeStereo', 'QueryStereo']
+__all__ = ['MoleculeStereo', 'Stereo']
